package com.sun.beans;

import java.util.Iterator;
import sun.reflect.generics.reflectiveObjects.GenericArrayTypeImpl;
import java.lang.reflect.Array;
import sun.reflect.generics.reflectiveObjects.ParameterizedTypeImpl;
import java.util.HashMap;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.GenericArrayType;
import java.util.Map;
import java.lang.reflect.Type;

public final class TypeResolver
{
    private static final WeakCache<Type, Map<Type, Type>> CACHE;
    
    public TypeResolver() {
    }
    
    public static Type resolveInClass(Class<?> clazz, Type type) {
        return resolve(getActualType(clazz), type);
    }
    
    public static Type[] resolveInClass(Class<?> clazz, Type[] array) {
        return resolve(getActualType(clazz), array);
    }
    
    public static Type resolve(Type type, Type obj) {
        if ((obj instanceof Class)) {
            return obj;
        }
        if ((obj instanceof GenericArrayType)) {
            Type resolve = resolve(type, ((GenericArrayType)obj).getGenericComponentType());
            return (resolve instanceof Class) ? Array.newInstance((Class)resolve, 0).getClass() : GenericArrayTypeImpl.make(resolve);
        }
        if ((obj instanceof ParameterizedType)) {
            ParameterizedType parameterizedType = (ParameterizedType)obj;
            return ParameterizedTypeImpl.make((Class)parameterizedType.getRawType(), resolve(type, parameterizedType.getActualTypeArguments()), parameterizedType.getOwnerType());
        }
        if ((obj instanceof WildcardType)) {
            WildcardType wildcardType = (WildcardType)obj;
            return new WildcardTypeImpl(resolve(type, wildcardType.getUpperBounds()), resolve(type, wildcardType.getLowerBounds()));
        }
        if ((obj instanceof TypeVariable)) {
            Map<Type, Type> map;
            synchronized (TypeResolver.CACHE) {
                map = (Map)TypeResolver.CACHE.get(type);
                if (map == null) {
                    map = new HashMap();
                    prepare(map, type);
                    TypeResolver.CACHE.put(type, map);
                }
            }
            Type type2 = (Type)map.get(obj);
            if (type2 == null || type2.equals(obj)) {
                return obj;
            }
            return resolve(type, fixGenericArray(type2));
        }
        else {
            throw new IllegalArgumentException("Bad Type kind: " + obj.getClass());
        }
    }
    
    public static Type[] resolve(Type type, Type[] array) {
        int length = array.length;
        Type[] array2 = new Type[length];
        for (int i = 0; i < length; ++i) {
            array2[i] = resolve(type, array[i]);
        }
        return array2;
    }
    
    public static Class<?> erase(Type type) {
        if ((type instanceof Class)) {
            return (Class)type;
        }
        if ((type instanceof ParameterizedType)) {
            return (Class)((ParameterizedType)type).getRawType();
        }
        if ((type instanceof TypeVariable)) {
            Type[] bounds = ((TypeVariable)type).getBounds();
            return (0 < bounds.length) ? erase(bounds[0]) : Object.class;
        }
        if ((type instanceof WildcardType)) {
            Type[] upperBounds = ((WildcardType)type).getUpperBounds();
            return (0 < upperBounds.length) ? erase(upperBounds[0]) : Object.class;
        }
        if ((type instanceof GenericArrayType)) {
            return Array.newInstance(erase(((GenericArrayType)type).getGenericComponentType()), 0).getClass();
        }
        throw new IllegalArgumentException("Unknown Type kind: " + type.getClass());
    }
    
    public static Class[] erase(Type[] array) {
        int length = array.length;
        Class[] array2 = new Class[length];
        for (int i = 0; i < length; ++i) {
            array2[i] = erase(array[i]);
        }
        return array2;
    }
    
    private static void prepare(Map<Type, Type> map, Type type) {
        Class<Object> clazz = (Class)(Class)((type instanceof Class) ? type : ((ParameterizedType)type).getRawType());
        TypeVariable[] typeParameters = clazz.getTypeParameters();
        Type[] array = (type instanceof Class) ? typeParameters : ((ParameterizedType)type).getActualTypeArguments();
        assert typeParameters.length == array.length;
        for (int i = 0; i < typeParameters.length; ++i) {
            map.put(typeParameters[i], array[i]);
        }
        Type genericSuperclass = clazz.getGenericSuperclass();
        if (genericSuperclass != null) {
            prepare(map, genericSuperclass);
        }
        Type[] genericInterfaces = clazz.getGenericInterfaces();
        for (int length = genericInterfaces.length, j = 0; j < length; ++j) {
            prepare(map, genericInterfaces[j]);
        }
        if (type instanceof Class && typeParameters.length > 0) {
            for (Map.Entry entry : map.entrySet()) {
                entry.setValue(erase((Type)entry.getValue()));
            }
        }
        return;
    }
    
    private static Type fixGenericArray(Type type) {
        if ((type instanceof GenericArrayType)) {
            Type fixGenericArray = fixGenericArray(((GenericArrayType)type).getGenericComponentType());
            if ((fixGenericArray instanceof Class)) {
                return Array.newInstance((Class)fixGenericArray, 0).getClass();
            }
        }
        return type;
    }
    
    private static Type getActualType(Class<?> rawType) {
        TypeVariable[] typeParameters = rawType.getTypeParameters();
        return (typeParameters.length == 0) ? rawType : ParameterizedTypeImpl.make(rawType, typeParameters, rawType.getEnclosingClass());
    }
    
    static {
        CACHE = new WeakCache();
    }
}
